import os
import re
import tempfile
import time
import asyncio
from typing import List
import datetime

from ruamel.yaml import YAML
import click

from conda_store import api, runner, utils, exception


async def parse_build(conda_store_api: api.CondaStoreAPI, uri: str):
    if re.fullmatch("(.+)/(.*):(.*)", uri):
        namespace, name, build_id = re.fullmatch("(.+)/(.*):(.*)", uri).groups()
        build = await conda_store_api.get_build(build_id)
        environment = await conda_store_api.get_environment(namespace, name)
        if build["environment_id"] != environment["id"]:
            raise exception.CondaStoreError(
                f"build {build_id} does not belong to environment {namespace}/{name}"
            )
        build_id = int(build_id)
    elif re.fullmatch("(.+)/(.*)", uri):
        namespace, name = re.fullmatch("(.+)/(.*)", uri).groups()
        environment = await conda_store_api.get_environment(namespace, name)
        build_id = environment["current_build_id"]
    if re.fullmatch(r"\d+", uri):  # build_id
        build_id = int(uri)

    return build_id


@click.group()
@click.option(
    "--conda-store-url",
    default="http://localhost:5000",
    envvar="CONDA_STORE_URL",
    help="conda-store base url including prefix",
)
@click.option(
    "--auth",
    envvar="CONDA_STORE_AUTH",
    type=click.Choice(["none", "token", "basic"], case_sensitive=False),
    help="conda-store authentication to use",
    default="none",
)
@click.option(
    "--no-verify-ssl",
    envvar="CONDA_STORE_NO_VERIFY",
    is_flag=True,
    default=False,
    help="Disable tls verification on API requests",
)
@click.pass_context
def cli(ctx, conda_store_url: str, auth: str, no_verify_ssl: bool):
    ctx.ensure_object(dict)
    ctx.obj["CONDA_STORE_API"] = api.CondaStoreAPI(
        conda_store_url=conda_store_url,
        verify_ssl=not no_verify_ssl,
        auth_type=auth,
    )


@cli.command(name="info")
@click.pass_context
@utils.coro
async def get_permissions(ctx):
    """Get current permissions and default namespace"""
    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        data = await conda_store.get_permissions()

    utils.console.print(
        f"Default namespace is [bold]{data['primary_namespace']}[/bold]"
    )
    utils.console.print(f"Authenticated [bold]{data['authenticated']}[/bold]")

    columns = {
        "Namespace": "namespace",
        "Name": "name",
        "Permissions": "permissions",
    }

    rows = []
    for key, value in data["entity_permissions"].items():
        namespace, name = key.split("/")
        rows.append(
            {
                "namespace": namespace,
                "name": name,
                "permissions": " ".join(_ for _ in value),
            }
        )

    utils.output_table("Permissions", columns, rows)


@cli.command(name="token")
@click.option("--namespace", help="Primary namespace to assign to new token")
@click.option("--expiration", type=int, help="Seconds from now for token to expire")
@click.option(
    "--permission",
    help="Permission to assign to token in form <namespace>/<name>:<role1>,<role2>,...",
    multiple=True,
)
@click.pass_context
@utils.coro
async def get_token(ctx, namespace: str, expiration: int, permission: List[str]):
    role_bindings = {}
    for p in permission:
        namespace_name, roles = p.split(":")
        roles = roles.split(",")
        role_bindings[namespace_name] = roles

    if expiration:
        expiration = str(
            datetime.datetime.now(tz=datetime.timezone.utc)
            + datetime.timedelta(seconds=expiration)
        )

    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        print(
            await conda_store.create_token(
                primary_namespace=namespace,
                expiration=expiration,
                role_bindings=role_bindings,
            ),
            end="",
        )


@cli.command(name="download")
@click.argument("uri")
@click.option(
    "--artifact",
    default="lockfile",
    type=click.Choice(["logs", "yaml", "lockfile", "archive"], case_sensitive=False),
)
@click.option(
    "--output-filename",
    help="Output filename for given download. build-{build_id}.{extension yaml|lock|tar.gz|image}",
)
@click.pass_context
@utils.coro
async def download(ctx, uri: str, artifact: str, output_filename: str = None):
    """Download artifacts for given build

    URI in format '<build-id>', '<namespace>/<name>', '<namespace>/<name>:<build-id>'
    """
    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        build_id = await parse_build(conda_store, uri)

        content = await conda_store.download(build_id, artifact)
        if output_filename is None:
            if artifact == "yaml":
                extension = "yaml"
            elif artifact == "lockfile":
                extension = "lock"
            elif artifact == "archive":
                extension = "tar.gz"
            elif artifact == "docker":
                extension = "image"
            output_filename = f"build-{build_id}.{extension}"

        with open(output_filename, "wb") as f:
            f.write(content)

        print(os.path.abspath(output_filename), end="")


@cli.command("wait")
@click.argument("uri")
@click.option(
    "--timeout",
    type=int,
    default=10 * 60,
    help="Time to wait for build to complete until reporting an error. Default 10 minutes",
)
@click.option(
    "--interval",
    type=int,
    default=10,
    help="Time to wait between polling for build status.Default 10 seconds",
)
@click.option(
    "--artifact",
    type=click.Choice(
        ["build", "lockfile", "yaml", "archive", "docker"], case_sensitive=False
    ),
    default="build",
    help="Choice of artifact to wait for. Default is 'build' which indicates the environment was built.",
)
@click.pass_context
@utils.coro
async def wait_environment(ctx, uri: str, timeout: int, interval: int, artifact: str):
    """Wait for given URI to complete or fail building

    URI in format '<build-id>', '<namespace>/<name>', '<namespace>/<name>:<build-id>'
    """
    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        build_id = await parse_build(conda_store, uri)

        start_time = time.time()
        while (time.time() - start_time) < timeout:
            build = await conda_store.get_build(build_id)
            build_artifact_types = set(
                _["artifact_type"] for _ in build["build_artifacts"]
            )

            if artifact == "build" and build["status"] == "COMPLETED":
                return
            elif artifact == "build" and build["status"] == "FAILED":
                raise exception.CondaStoreError(f"Build {build_id} failed")
            elif artifact == "lockfile" and "LOCKFILE" in build_artifact_types:
                return
            elif artifact == "yaml" and "YAML" in build_artifact_types:
                return
            elif artifact == "archive" and "CONDA_PACK" in build_artifact_types:
                return
            elif artifact == "docker" and "DOCKER_MANIFEST" in build_artifact_types:
                return
            await asyncio.sleep(interval)

        raise exception.CondaStoreError(
            f"Build {build_id} failed to complete in {timeout} seconds"
        )


@cli.command("run")
@click.option(
    "--artifact",
    default="archive",
    type=click.Choice(["yaml", "lockfile", "archive"], case_sensitive=False),
    help="Artifact type to use for execution. Conda-Pack is the default format",
)
@click.option(
    "--no-cache",
    is_flag=True,
    default=False,
    help="Disable caching builds for fast execution",
)
@click.argument("uri")
@click.argument("command", nargs=-1)
@click.pass_context
@utils.coro
async def run_environment(ctx, uri: str, no_cache: bool, command: str, artifact: str):
    """Execute given environment specified as a URI with COMMAND

    URI in format '<build-id>', '<namespace>/<name>', '<namespace>/<name>:<build-id>'\n
    COMMAND is a list of arguments to execute in given environment
    """
    if len(command) == 0:
        command = ["python"]

    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        build_id = await parse_build(conda_store, uri)

        if no_cache:
            with tempfile.TemporaryDirectory() as tmpdir:
                await runner.run_build(conda_store, tmpdir, build_id, command, artifact)
        else:
            directory = os.path.join(
                tempfile.gettempdir(), "conda-store", str(build_id)
            )
            os.makedirs(directory, exist_ok=True)
            await runner.run_build(conda_store, directory, build_id, command, artifact)


@cli.command("solve")
@click.argument("filename", type=click.File("r"))
@click.option(
    "--output",
    default="table",
    type=click.Choice(["json", "table"]),
    help="Output format to display solve. Default table.",
)
@click.pass_context
@utils.coro
async def solve_environment(ctx, filename, output: str):
    """Remotely solve given environment.yaml"""
    yaml = YAML(typ="safe")
    data = yaml.load(filename)

    def _get_pip(data):
        for _ in data.get("dependencies", []):
            if isinstance(_, dict):
                return _.get("pip", [])
        return []

    arguments = {
        "channels": data.get("channels", []),
        "conda": [_ for _ in data.get("dependencies", []) if isinstance(_, str)],
        "pip": _get_pip(data),
    }

    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        packages = await conda_store.solve_environment(**arguments)

    total_size = 0
    for package in packages:
        total_size += package["size"]
        package["size"] = utils.sizeof_fmt(package["size"])

    if output == "table":
        utils.output_table(
            "Packages",
            {
                "Name": "name",
                "Version": "version",
                "Channel": "channel_id",
                "Size": "size",
            },
            packages,
        )
        utils.console.print(f"Total Size: {utils.sizeof_fmt(total_size)}")
    elif output == "json":
        utils.output_json(packages)


# ================= LIST ==================
@cli.group("list")
def list_group():
    pass


@list_group.command("namespace")
@click.option(
    "--output",
    default="table",
    type=click.Choice(["json", "table"]),
    help="Output format to display builds. Default table.",
)
@click.pass_context
@utils.coro
async def list_namespace(ctx, output: str):
    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        namespaces = await conda_store.list_namespaces()

    if output == "table":
        utils.output_table("Namespaces", {"Id": "id", "Name": "name"}, namespaces)
    elif output == "json":
        utils.output_json(namespaces)


@list_group.command("build")
@click.option(
    "--status",
    # see conda_store_server.schema.BuildStatus
    type=click.Choice(["QUEUED", "BUILDING", "COMPLETED", "FAILED"]),
    help="Filter builds which have given status",
)
@click.option(
    "--artifact",
    # see conda_store_server.schema.BuildArtifactType
    type=click.Choice(
        [
            "DIRECTORY",
            "LOCKFILE",
            "LOGS",
            "YAML",
            "CONDA_PACK",
            "DOCKER_BLOB",
            "DOCKER_MANIFEST",
        ]
    ),
    help="Filter builds which have given artifact",
)
@click.option(
    "--package",
    help="Filter builds which have given package (can be used multiple times)",
    multiple=True,
)
@click.option(
    "--output",
    default="table",
    type=click.Choice(["json", "table"]),
    help="Output format to display builds. Default table.",
)
@click.pass_context
@utils.coro
async def list_build(ctx, output: str, status: str, artifact: str, package: List[str]):
    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        builds = await conda_store.list_builds(
            status=status,
            artifact=artifact,
            packages=package,
        )

    for build in builds:
        build["size"] = utils.sizeof_fmt(build["size"])

    if output == "table":
        utils.output_table(
            "Builds", {"Id": "id", "Size": "size", "Status": "status"}, builds
        )
    elif output == "json":
        utils.output_json(builds)


@list_group.command("environment")
@click.option(
    "--status",
    # see conda_store_server.schema.BuildStatus
    type=click.Choice(["QUEUED", "BUILDING", "COMPLETED", "FAILED"]),
    help="Filter environments which have given status",
)
@click.option(
    "--artifact",
    # see conda_store_server.schema.BuildArtifactType
    type=click.Choice(
        [
            "DIRECTORY",
            "LOCKFILE",
            "LOGS",
            "YAML",
            "CONDA_PACK",
            "DOCKER_BLOB",
            "DOCKER_MANIFEST",
        ]
    ),
    help="Filter environments which have given artifact",
)
@click.option(
    "--package",
    help="Filter environments which have given package (can be used multiple times)",
    multiple=True,
)
@click.option(
    "--output",
    default="table",
    type=click.Choice(["json", "table"]),
    help="Output format to display builds. Default table.",
)
@click.pass_context
@utils.coro
async def list_environment(
    ctx, output: str, status: str, artifact: str, package: List[str]
):
    async with ctx.obj["CONDA_STORE_API"] as conda_store:
        builds = await conda_store.list_environments(
            status=status,
            artifact=artifact,
            packages=package,
        )

    if output == "table":
        utils.output_table(
            "Environments",
            {"Id": "id", "Namespace Id": "namespace.id", "Name": "name"},
            builds,
        )
    elif output == "json":
        utils.output_json(builds)



from typing import List
import os
import io
import tarfile
import subprocess
import shlex

from conda_store.exception import CondaStoreError
from conda_store import api


async def run_build(
    conda_store_api: api.CondaStoreAPI,
    directory: str,
    build_id: int,
    command: List[str],
    artifact="archive",
):
    if artifact == "archive":
        await run_build_archive(conda_store_api, directory, build_id, command)
    else:
        raise CondaStoreError(f"Running build artifact {artifact} not supported")


async def run_build_archive(
    conda_store_api: api.CondaStoreAPI,
    conda_prefix: str,
    build_id: int,
    command: List[str],
):
    activate = os.path.join(conda_prefix, "bin", "activate")
    conda_unpack = os.path.join(conda_prefix, "bin", "conda-unpack")

    if not os.path.isfile(activate):
        content = await conda_store_api.download(build_id, "archive")
        content = io.BytesIO(content)

        tarfile.open(fileobj=content, mode="r:gz").extractall(path=conda_prefix)

        if os.path.exists(conda_unpack):
            subprocess.check_output(conda_unpack)

    wrapped_command = [
        "bash",
        "-c",
        ". '{}' '{}' && exec {}".format(
            activate, conda_prefix, " ".join(shlex.quote(c) for c in command)
        ),
    ]
    os.execvp(wrapped_command[0], wrapped_command)



from typing import List, Dict
import asyncio
import sys
import functools
import json

from rich.console import Console
from rich.table import Table

from conda_store import exception


console = Console()
error_console = Console(stderr=True, style="bold red")


def coro(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return asyncio.run(f(*args, **kwargs))
        except exception.CondaStoreError as e:
            error_console.print(e.args[0])
            sys.exit(1)

    return wrapper


def flatten(d: Dict):
    _d = {}
    for key, value in d.items():
        if isinstance(value, dict):
            for _key, _value in flatten(value).items():
                _d[f"{key}.{_key}"] = _value
        else:
            _d[key] = value
    return _d


def lookup(d: Dict, key: str):
    _d = d
    keys = key.split(".")
    for key in keys:
        _d = _d[key]
    return _d


def sizeof_fmt(num, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def timedelta_fmt(td):
    """
    Returns a humanized string representing timedelta
    """

    def plural(unit, word):
        if unit > 1:
            return word + "s"
        return word

    years = td.days // 365
    months = td.days // 30
    days = td.days
    hours = td.seconds / 3600
    minutes = td.seconds % 60
    seconds = td.seconds

    if years > 0:
        return f"{years} {plural(years, 'year')}"
    elif months > 0:
        return f"{months} {plural(months, 'month')}"
    elif days > 0:
        return f"{days} {plural(days, 'day')}"
    elif hours > 0:
        return f"{hours} {plural(hours, 'hour')}"
    elif minutes > 0:
        return f"{minutes} {plural(minutes, 'minute')}"
    elif seconds > 0:
        return f"{seconds} {plural(seconds, 'second')}"


def output_json(data, **kwargs):
    print(json.dumps(data, **kwargs), end="")


def output_table(title: str, columns: Dict[str, str], rows: List[Dict]):
    table = Table(title=title)
    for column in columns.keys():
        table.add_column(column)

    for row in rows:
        table.add_row(*[str(lookup(row, key)) for key in columns.values()])

    console.print(table)



import aiohttp
import yarl


async def none_authentication(verify_ssl: bool = True):
    return aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(ssl=None if verify_ssl else False),
    )


async def token_authentication(api_token: str, verify_ssl: bool = True):
    return aiohttp.ClientSession(
        headers={"Authorization": f"token {api_token}"},
        connector=aiohttp.TCPConnector(ssl=None if verify_ssl else False),
    )


async def basic_authentication(
    conda_store_url, username, password, verify_ssl: bool = True
):
    session = aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(ssl=None if verify_ssl else False),
    )

    await session.post(
        yarl.URL(conda_store_url) / "login",
        json={
            "username": username,
            "password": password,
        },
    )

    return session



import os
import math
import datetime
from typing import List, Dict

import yarl

from conda_store import auth, exception


class CondaStoreAPIError(exception.CondaStoreError):
    pass


class CondaStoreAPI:
    def __init__(
        self, conda_store_url: str, auth_type: str = "none", verify_ssl=True, **kwargs
    ):
        self.conda_store_url = yarl.URL(conda_store_url)
        self.api_url = self.conda_store_url / "api/v1"
        self.auth_type = auth_type
        self.verify_ssl = verify_ssl

        if auth_type == "token":
            self.api_token = kwargs.get("api_token", os.environ["CONDA_STORE_TOKEN"])
        elif auth_type == "basic":
            self.username = kwargs.get("username", os.environ["CONDA_STORE_USERNAME"])
            self.password = kwargs.get("password", os.environ["CONDA_STORE_PASSWORD"])

    async def __aenter__(self):
        if self.auth_type == "none":
            self.session = await auth.none_authentication(verify_ssl=self.verify_ssl)
        elif self.auth_type == "token":
            self.session = await auth.token_authentication(
                self.api_token, verify_ssl=self.verify_ssl
            )
        elif self.auth_type == "basic":
            self.session = await auth.basic_authentication(
                self.conda_store_url,
                self.username,
                self.password,
                verify_ssl=self.verify_ssl,
            )
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()

    async def get_paginated_request(self, url: yarl.URL, max_pages=None, **kwargs):
        data = []

        async with self.session.get(url) as response:
            response_data = await response.json()
            num_pages = math.ceil(response_data["count"] / response_data["size"])
            data.extend(response_data["data"])

        if max_pages is not None:
            num_pages = min(max_pages, num_pages)

        for page in range(2, num_pages + 1):
            async with self.session.get(url % {"page": page}) as response:
                data.extend((await response.json())["data"])

        return data

    async def get_permissions(self):
        async with self.session.get(self.api_url / "permission") as response:
            return (await response.json())["data"]

    async def create_token(
        self,
        primary_namespace: str = None,
        role_bindings: Dict[str, List[str]] = None,
        expiration: datetime.datetime = None,
    ):
        current_permissions = await self.get_permissions()
        requested_permissions = {
            "primary_namespace": primary_namespace
            or current_permissions["primary_namespace"],
            "role_bindings": role_bindings or current_permissions["entity_roles"],
            "exp": expiration or current_permissions["expiration"],
        }
        async with self.session.post(
            self.api_url / "token", json=requested_permissions
        ) as response:
            if response.status == 400:
                raise CondaStoreAPIError((await response.json())["message"])

            return (await response.json())["data"]["token"]

    async def list_namespaces(self):
        return await self.get_paginated_request(self.api_url / "namespace")

    async def create_namespace(self, namespace: str):
        async with self.session.post(
            self.api_url / "namespace" / namespace
        ) as response:
            if response.status != 200:
                raise CondaStoreAPIError(f"Error creating namespace {namespace}")

    async def delete_namespace(self, namespace: str):
        async with self.session.delete(
            self.api_url / "namespace" / namespace
        ) as response:
            if response.status != 200:
                raise CondaStoreAPIError(f"Error deleting namespace {namespace}")

    async def list_environments(self, status: str, artifact: str, packages: List[str]):
        url = self.api_url / "environment"
        if status:
            url = url % {"status": status}
        if artifact:
            url = url % {"artifact": artifact}
        if packages:
            url = url % {"packages": packages}
        return await self.get_paginated_request(url)

    async def delete_environment(self, namespace: str, name: str):
        async with self.session.delete(
            self.api_url / "environment" / namespace / name
        ) as response:
            if response.status != 200:
                raise CondaStoreAPIError(
                    f"Error deleting environment {namespace}/{name}"
                )

    async def create_environment(self, namespace: str, specification: str):
        async with self.session.post(
            self.api_url / "specification",
            json={
                "namespace": namespace,
                "specification": specification,
            },
        ) as response:
            data = await response.json()
            if response.status != 200:
                message = data["message"]
                raise CondaStoreAPIError(
                    f"Error creating environment in namespace {namespace}\nReason {message}"
                )

            return data["data"]["build_id"]

    async def get_environment(self, namespace: str, name: str):
        async with self.session.get(
            self.api_url / "environment" / namespace / name
        ) as response:
            if response.status != 200:
                raise CondaStoreAPIError(
                    f"Error getting environment {namespace}/{name}"
                )

            return (await response.json())["data"]

    async def solve_environment(
        self, channels: List[str], conda: List[str], pip: List[str]
    ):
        async with self.session.get(
            self.api_url
            / "specification"
            % {
                "channels": channels,
                "conda": conda,
                "pip": pip,
            }
        ) as response:
            return (await response.json())["solve"]

    async def list_builds(self, status: str, artifact: str, packages: List[str]):
        url = self.api_url / "build"
        if status:
            url = url % {"status": status}
        if artifact:
            url = url % {"artifact": artifact}
        if packages:
            url = url % {"packages": packages}
        return await self.get_paginated_request(url)

    async def get_build(self, build_id: int):
        async with self.session.get(self.api_url / "build" / str(build_id)) as response:
            if response.status != 200:
                raise CondaStoreAPIError(f"Error getting build {build_id}")

            return (await response.json())["data"]

    async def download(self, build_id: int, artifact: str) -> bytes:
        url = self.api_url / "build" / str(build_id) / artifact
        async with self.session.get(url) as response:
            if response.status != 200:
                raise CondaStoreAPIError(f"Error downloading build {build_id}")

            return await response.content.read()
